unit RKSIni;
{===========================================================================================
 TRKSIniFile class ver 1.02   Copyright 1996..1999 by
        Roman Krejci-RKS,
        Pujmanove 1552,
 140 00 Prague 4,
        Czech Republic
 http://www.rksolution.cz, EMail INFO@rksolution.cz

 TRKSIniFile replaces the standard Delphi VCL tIniFile.

 Instead of using WIN API Get/Write(Private)ProfileXXXX calls, it treates the INI file
 as an ordinary text file. This avoids problems caused by inconsistencies in the API
 behaviour between Win9x/NT. Also - as Microsoft constantly pushes for
 the INI-to-REGISTRY switch, it sounds like a good idea to completely abandon
 (Private)Profile API anyway.

 Upon object creation, the class reads the entire file content into private
 stringlists. All subsequent writes do affect these stringlists only.
 To update the disk file content, it is necessary to:
  1) call FLUSH method (not present in Delphi tIniFile) either directly
     or indirectly (WriteString with all arguments empty)
  or
  2) free all objects associated with given disk file - within its destructor,
     the class calls FLUSH if there were any changes to the INI content.

 The TRKSIniFile delegates all methods to another (hidden) class - TUniqueIniFile.
 There is only one instance of TUniqueIniFile per physical INI file. This way
 several independent instances of TRKSIniFile do all share the same actual
 INI file content.

 WARNINGS! No trim is performed, so what you write, you get.
           Do not access the same INI file by both TRKSIniFile and TINIfile class.

 Vesrion history
 1.00      December 1999         initial release
 1.01      January   4 2000      some initial bugs fixed
 1.02      February 21 2000      fixed "stack overflow" bug in 32-bit version
                                 when the INI file is located on remote drive
 ===========================================================================================}
{$IFDEF WIN32}
  {$H+}
{$ELSE}
  {$S-}
{$ENDIF}
  {$B-}
  {$I+}
  {$V-}
  {$X+}
interface
uses
{$IFDEF WIN32}
  Windows,
{$ELSE}
  WinTypes, WinProcs,
{$ENDIF}
  SysUtils, Classes;

type
  TRKSIniFile = class(TObject)
  private
    fUniqueINI: tObject; { object that actually performs all operations }
    function    GetFileName: string;
  public
    constructor Create(const aFileName: string);
    destructor Destroy; override;
{the following routine (Flush) is not present in Delphi tINIFile: }
    procedure Flush;
    function ReadString(const Section, Ident, Default: string): string;
    procedure WriteString(const Section, Ident, Value: String);
    function ReadInteger(const Section, Ident: string;
      Default: Longint): Longint;
    procedure WriteInteger(const Section, Ident: string; Value: Longint);
    function ReadBool(const Section, Ident: string;
      Default: Boolean): Boolean;
    procedure WriteBool(const Section, Ident: string; Value: Boolean);
    procedure ReadSection(const Section: string; Strings: TStrings);
    procedure ReadSections(Strings: TStrings);
    procedure ReadSectionValues(const Section: string; Strings: TStrings);
    procedure EraseSection(const Section: string);
    procedure DeleteKey(const Section, Ident: String);
    property  FileName: string read GetFileName;
  end;

{$IFDEF WIN32}
{ For an existing file or directory, the function returns
  long path version. If the file or directory does not exist,
  the function returns the Fname argument. }
function GetTrueLongFilename(Fname: string): string;
{$ENDIF}

implementation

type
  TUniqueIniFile = class(TObject)  { Only one instance per physical INI file }
  private
    FFileName: string;
    LeadingText,
    FSEctions: tStringList;
    fChanged : boolean;
    fRefCount : integer;
    Function FindSectionValues(const Section: string): tStringList;
    Function GetSectionValues(const Section: string): tStringList;
    procedure SetRefCount(Value: Integer);
  public
    constructor Create(const aFileName: string);
    destructor Destroy; override;
    procedure Flush;
    function ReadString(const Section, Ident, Default: string): string;
    procedure WriteString(const Section, Ident, Value: String);
    function ReadInteger(const Section, Ident: string;
      Default: Longint): Longint;
    procedure WriteInteger(const Section, Ident: string; Value: Longint);
    function ReadBool(const Section, Ident: string;
      Default: Boolean): Boolean;
    procedure WriteBool(const Section, Ident: string; Value: Boolean);
    procedure ReadSection(const Section: string; Strings: TStrings);
    procedure ReadSections(Strings: TStrings);
    procedure ReadSectionValues(const Section: string; Strings: TStrings);
    procedure EraseSection(const Section: string);
    procedure DeleteKey(const Section, Ident: String);
    property FileName: string read FFileName;
    property RefCount: integer read fRefCount write SetRefCount;
  end;

{ helpers }

{$IFNDEF WIN32}
const
  max_Path = 176;

function ExtractFileDir(const FileName: string): string;
var
  I: Integer;
begin
  I := Length(FileName);
  while (I > 0) and not (FileName[I] in ['\', ':']) do Dec(I);
  if (I > 1) and (FileName[I] = '\') and
    not (FileName[I - 1] in ['\', ':']) then Dec(I);
  Result := Copy(FileName, 1, I);
end;
{$ENDIF}

{$IFDEF WIN32}
function GetTrueLongFilename(Fname: string): string;
var
  RootNameLen,
  IterCount: integer;

    function GetLFilename(const Path: string): string; {called recursively}
    var SR: tSearchRec;
        FindResult: integer;
        Fname: string;
    begin
      Inc(IterCount);
    { ver 1.02 - check for Root directory part }
      if length(path) <= RootNameLen then Result := path else begin
        FindResult := FindFirst(path,faAnyFile and not faVolumeID,SR);
        if FindResult = 0 then begin
          Fname := StrPas(SR.Finddata.cFileName);
          FindClose(SR);
          result := GetLFilename(ExtractFileDir(Path))+Fname;
        end else { path does not exist }
          result := Path;
      End;
      Dec(IterCount);
      if (IterCount > 0) and (Result <>'') and (result[length(Result)] <> '\') then
        Result := Result + '\';
    end;
Var
  FnameLen: integer;
  AdjustBackslash: boolean;
  fDrive: String;
begin
  IterCount := 0;
  fDrive := ExtractFileDrive(Fname);
  RootNameLen := length(fDrive) + 1;
  Fnamelen := length(Fname);
  if (FnameLen > 0) and (Fname[FnameLen]='\') then begin
    AdjustBackslash := True;
    Fname := Copy(Fname,1,FnameLen-1);
  end else
    AdjustBackSlash := false;
  Result := GetLFilename(Fname);
  if AdjustBackslash then Result := Result + '\';
end;
{$ENDIF}

Var
  UniqueINIList: tList;

{ UniqueINIList management }
function GetUniqueINI(const FileName: string): TUniqueIniFile;
VAR WinPath: array[0..max_path] of Char;
    ActualFileDir,ActualFilename: string;
    I,Len: integer;
begin
  Result := NIL;
  if Filename <> '' then
    ActualFilename := Filename
  else
    ActualFilename := 'WIN.INI';
  ActualFileDir  := ExtractFileDir(ActualFileName);
  if ActualFileDir = '' then {search in WINDOWS folder} begin
    len := GetWindowsDirectory(WinPath,max_path);
    if (len > 0) and (len <= max_path) then begin
      ActualFileDir := StrPas(WinPath);
      if not (ActualFileDir[len] in ['\',':',#0]) then
        ActualFileDir := ActualFileDir + '\';
      ActualFilename := ActualFileDir + ActualFilename;
    end
      else
        raise exception.Create('Unable to retrieve Windows directory');
  end;
  ActualFilename := ExpandFilename(ActualFilename);
{$IFDEF WIN32}
  if FileExists(ActualFilename) then
    ActualFilename := GetTrueLongFilename(ActualFilename)
  else
    ActualFilename := GetTrueLongFilename(ExtractFilePath(ActualFileName))+
                            ExtractFileName(ActualFileName);
{$ENDIF}
  if UniqueINIList = NIL then UniqueINIList := tList.Create
  else
  For i := 0 to UniqueINIList.Count-1 do
    if CompareText(TUniqueIniFile(UniqueINIList[i]).Filename,ActualFilename)=0
      then begin
        Result := TUniqueIniFile(UniqueINIList[i]);
        break;
      end;
  if Result = nil then begin
    Result := TUniqueIniFile.Create(ActualFilename);
    Try
      UniqueINIList.Add(Result);
    Except
      Result.Free;
      raise;
    End;
  end;
end;

{ TUniqueIniFile }

Function TUniqueIniFile.FindSectionValues(const Section: string): tStringList;
Var i: integer;
begin
  I := fSections.IndexOf(Section); { bug fix in 1.01 }
  if I >= 0 then
    Result := tStringList(fSections.Objects[i])
  else
    Result := nil;
end;

Function TUniqueIniFile.GetSectionValues(const Section: string): tStringList;
begin
  Result := FindSectionValues(Section);
  if Result = nil then begin { add section }
    Result := tStringList.Create;
    Result.duplicates := dupIgnore;
    Try
      FSEctions.AddObject(Section,Result);
    Except
      Result.Free;
      raise;
    End;
  End;
End;

constructor TUniqueIniFile.Create(const aFileName: string);
var txt: system.text;
    SectValues: tStringList;
    Line: string;
    Len: integer;
begin
  inherited Create;
  FSEctions := tStringList.Create;
  LeadingText := tStringList.Create;
  FSEctions.duplicates := dupIgnore;
  if FileExists(aFilename) then begin
    SectValues := nil;
    assignFile(txt,aFilename);
    Reset(txt);
    Try
      While not EOF(txt) do begin
        ReadLn(Txt,Line);
        Len  := length(Line);
        if Len > 0 then begin
          if (Line[1]='[') and (Line[len]=']') then begin
            Line := Copy(Line,2,len-2);
            SectValues := GetSectionValues(Line);
            continue; { skips the rest of the loop }
          end;
          if SectValues <> nil then SectValues.Add(Line)
        end;
        if SectValues = nil then LeadingText.Add(Line)
      end;
    Finally
      CloseFile(txt);
    End;
  End;
  fFilename := aFilename;
end;

destructor TUniqueIniFile.Destroy;
var i: integer;
begin
  if fChanged then Flush;
  LeadingText.Free;
  for i := 0 to FSEctions.Count-1 do
    FSEctions.Objects[i].Free;
  FSEctions.Free;
  inherited Destroy;
end;

procedure TUniqueIniFile.Flush;
var txt: system.text;
    i,j: integer;
    SectValues: tStringList;
begin
  if fFilename <> '' then begin
    assignFile(txt,fFilename);
    Rewrite(txt);
    Try
      for i := 0 to LeadingText.Count-1 do
        Writeln(Txt,LeadingText[i]);
      for i := 0 to FSEctions.Count-1 do begin
        if i > 0 then Writeln(txt);
        Writeln(Txt,Format('[%s]',[FSEctions[i]]));
        SectValues := tStringList(FSEctions.Objects[i]);
        for j := 0 to SectValues.Count-1 do
          Writeln(Txt,SectValues[j]);
      end;
    Finally
      CloseFile(Txt);
    End;
  end;
  fChanged := false;
end;

function TUniqueIniFile.ReadString(const Section, Ident, Default: string): string;
Var SectValues: tStringList;
begin
  SectValues := FindSectionValues(Section);
  if SectValues <> nil then begin
    Result := SectValues.Values[Ident];
    if Result = '' then Result := Default;
  end
    else
      Result := Default;
end;

procedure TUniqueIniFile.WriteString(const Section, Ident, Value: String);
Var SectValues: tStringList;
    i: integer;
begin
  if (Section = '') and (Ident='') and (Value='') then Flush else { flushes onto disk }
  if Ident = '' then begin { erases section }
    SectValues := FindSectionValues(Section);
    if SectValues <> nil then begin
      i := fSections.IndexOfObject(SectValues);
      fSections.Delete(i);
      SectValues.Free;
      fChanged := true;
    end
  end else begin
    SectValues := GetSectionValues(Section);
    SectValues.Values[Ident] := Value;
    fChanged := true;
  end;
end;


function TUniqueIniFile.ReadInteger(const Section, Ident: string;
  Default: Longint): Longint;
Var s: string;
    code: integer;
begin
  s := ReadString(Section,Ident,'');
  if s = '' then Result := Default else begin
    if (Length(S) > 2) and (S[1] = '0') and (S[2] in ['X','x']) then
      S := '$' + Copy(S, 3, Maxint);
    val(s,result,code);
    if code <> 0 then Result := default;
  end;
end;

procedure TUniqueIniFile.WriteInteger(const Section, Ident: string; Value: Longint);
begin
  WriteString(Section, Ident, IntToStr(Value));
end;

function TUniqueIniFile.ReadBool(const Section, Ident: string;
  Default: Boolean): Boolean;
begin
  Result := ReadInteger(Section, Ident, Ord(Default)) <> 0;
end;

procedure TUniqueIniFile.WriteBool(const Section, Ident: string; Value: Boolean);
begin
  WriteInteger(Section, Ident, ord(value));
end;

procedure TUniqueIniFile.ReadSection(const Section: string; Strings: TStrings);
var SectValues: tStringList;
    i: integer;
{$IFNDEF WIN32}
    s: string;
    P: integer;
{$ENDIF}
begin
  Strings.Clear;
  SectValues := FindSectionValues(Section);
  if SectValues <> nil then begin
    Strings.BeginUpdate;
    try
      For i := 0 to SectValues.Count-1 do
{$IFNDEF WIN32}
        begin
          s := SectValues[i];
          P := Pos('=', s);
          if P > 0 then
            s := copy(s,1, P-1)
          else
            s := '';
          Strings.Add(s)
        end;

{$ELSE}
        Strings.Add(SectValues.names[i])
{$ENDIF}
    finally
      Strings.EndUpdate;
    End;
  end;
end;

procedure TUniqueIniFile.ReadSections(Strings: TStrings);
var i: integer;
begin
  Strings.Clear;
  Strings.BeginUpdate;
  try
    For i := 0 to fSections.Count-1 do
      Strings.Add(fSections[i]);
  finally
    Strings.EndUpdate;
  end;
end;

procedure TUniqueIniFile.ReadSectionValues(const Section: string; Strings: TStrings);
var SectValues: tStringList;
begin
  Strings.Clear;
  SectValues := FindSectionValues(Section);
  if SectValues <> nil then
    Strings.Assign(SectValues);
end;

procedure TUniqueIniFile.EraseSection(const Section: string);
begin
  WriteString(Section,'','');
end;

procedure TUniqueIniFile.DeleteKey(const Section, Ident: String);
begin
  WriteString(Section,Ident,'');
end;

procedure TUniqueIniFile.SetRefCount(Value: Integer);
var i: integer;
begin
  if Value <= 0 then begin
    i := UniqueINIList.IndexOf(Self);
    if i >= 0 then UniqueINIList.Delete(i);
    if UniqueINIList.Count = 0 then begin
      UniqueINIList.Free;
      UniqueINIList := NIL;
    end;
    Free;
  end else fRefCount := value;
end;

{ TRKSIniFile }

constructor TRKSIniFile.Create(const aFileName: string);
begin
  fUniqueINI := GetUniqueINI(aFileName);
  With TUniqueIniFile(fUniqueINI) do
    RefCount := RefCount + 1;
end;

destructor  TRKSIniFile.Destroy;
Var UniqueINI: TUniqueIniFile;
begin
  UniqueINI  := TUniqueIniFile(fUniqueINI);
  fUniqueINI := NIL;
  if UniqueINI <> nil then UniqueINI.RefCount := UniqueINI.RefCount - 1;
  inherited Destroy;
end;

procedure   TRKSIniFile.Flush;
begin
  TUniqueIniFile(fUniqueINI).Flush;
end;

function    TRKSIniFile.ReadString(const Section, Ident, Default: string): string;
begin
  Result := TUniqueIniFile(fUniqueINI).ReadString(Section, Ident, Default);
end;

procedure   TRKSIniFile.WriteString(const Section, Ident, Value: String);
begin
  TUniqueIniFile(fUniqueINI).WriteString(Section, Ident, Value);
end;

function    TRKSIniFile.ReadInteger(const Section, Ident: string;
  Default: Longint): Longint;
begin
  Result := TUniqueIniFile(fUniqueINI).ReadInteger(Section, Ident, Default);
end;

procedure   TRKSIniFile.WriteInteger(const Section, Ident: string; Value: Longint);
begin
  TUniqueIniFile(fUniqueINI).WriteInteger(Section, Ident, Value);
end;

function    TRKSIniFile.ReadBool(const Section, Ident: string;
  Default: Boolean): Boolean;
begin
  Result := TUniqueIniFile(fUniqueINI).ReadBool(Section, Ident, Default);
end;

procedure   TRKSIniFile.WriteBool(const Section, Ident: string; Value: Boolean);
begin
  TUniqueIniFile(fUniqueINI).WriteBool(Section, Ident, Value);
end;

procedure   TRKSIniFile.ReadSection(const Section: string; Strings: TStrings);
begin
  TUniqueIniFile(fUniqueINI).ReadSection(Section, Strings);
end;

procedure   TRKSIniFile.ReadSections(Strings: TStrings);
begin
  TUniqueIniFile(fUniqueINI).ReadSections(Strings);
end;

procedure   TRKSIniFile.ReadSectionValues(const Section: string; Strings: TStrings);
begin
  TUniqueIniFile(fUniqueINI).ReadSectionValues(Section,Strings);
end;

procedure   TRKSIniFile.EraseSection(const Section: string);
begin
  TUniqueIniFile(fUniqueINI).EraseSection(Section);
end;

procedure   TRKSIniFile.DeleteKey(const Section, Ident: String);
begin
  TUniqueIniFile(fUniqueINI).DeleteKey(Section, Ident);
end;

function    TRKSIniFile.GetFileName: string;
begin
  Result := TUniqueIniFile(fUniqueINI).Filename;
end;



end.
